from __future__ import print_function
import numpy as np
from sympy import Symbol,Function
import sympy
from pixell import fft as efft, enmap
import os,sys
from . import utils

"""
Routines to reduce and evaluate symbolic mode coupling integrals
"""

# Built-in special symbols
l1x = Symbol('l1x') # \vec{l}_1x
l1y = Symbol('l1y') # \vec{l}_1y
l2x = Symbol('l2x') # \vec{l}_2x
l2y = Symbol('l2y') # \vec{l}_2y
l1 = Symbol('l1') # |\vec{l}_1|
l2 = Symbol('l2') # |\vec{l}_2|
Lx = Symbol('Lx') # \vec{L}_x
Ly = Symbol('Ly') # \vec{L}_y
L = Symbol('L') # |\vec{L}|
Ldl1 = (Lx*l1x+Ly*l1y) # \vec{L}.\vec{l}_1
Ldl2 = (Lx*l2x+Ly*l2y) # \vec{L}.\vec{l}_2
Lxl1 = (Ly*l1x-Lx*l1y) # \vec{L} x \vec{l}_1 for curl
Lxl2 = (Ly*l2x-Lx*l2y) # \vec{L} x \vec{l}_2 for curl

# More built-in special symbols
# cos(2\theta_{12}), sin(2\theta_{12}) for polarization
cos2t12,sin2t12 = utils.substitute_trig(l1x,l1y,l2x,l2y,l1,l2)

# Custom symbol wrapper
def e(symbol):
    # TODO: add exceptions if symbol doesn't correspond to key structure
    return Symbol(symbol)

ifft = lambda x: enmap.ifft(x,normalize='phys')
fft = lambda x: enmap.fft(x,normalize='phys')

evaluate = utils.evaluate

def factorize_2d_convolution_integral(expr,l1funcs=None,l2funcs=None,groups=None,validate=True):
    """Reduce a sympy expression of variables l1x,l1y,l2x,l2y,l1,l2 into a sum of 
    products of factors that depend only on vec(l1) and vec(l2) and neither, each. If the expression
    appeared as the integrand in an integral over vec(l1), where 
    vec(l2) = vec(L) - vec(l1) then this reduction allows one to evaluate the 
    integral as a function of vec(L) using FFTs instead of as a convolution.

    Parameters
    ----------

    expr: :obj:`sympy.core.symbol.Symbol` 
        The full Sympy expression to reduce to sum of products of functions of l1 and l2.

    l1funcs: list
        List of symbols that are functions of l1
    l2funcs: list
        List of symbols that are functions of l2
    groups: list,optional 
        Group all terms such that they have common factors of the provided list of 
        expressions to reduce the number of FFTs.
    validate: boolean,optional
        Whether to check that the final expression and the original agree. Defaults to True.

    Returns
    -------

    terms
    unique_l1s
    unique_l2s
    ogroups
    ogroup_weights
    ogroup_symbols

    """

    # Generic message if validation fails
    val_fail_message = "Validation failed. This expression is likely not reducible to FFT form."
    # Get the 2D convolution cartesian variables
    # l1x,l1y,l2x,l2y,l1,l2 = get_ells()
    if l1funcs is None: l1funcs = []
    if l2funcs is None: l2funcs = []
    if l1x not in l1funcs: l1funcs.append(l1x)
    if l1y not in l1funcs: l1funcs.append(l1y) 
    if l1 not in l1funcs: l1funcs.append(l1)
    if l2x not in l2funcs: l2funcs.append(l2x)
    if l2y not in l2funcs: l2funcs.append(l2y) 
    if l2 not in l2funcs: l2funcs.append(l2)
    Lx = Symbol('Lx')
    Ly = Symbol('Ly')
    L = Symbol('L')
    ofuncs1 = set(l1funcs) - set([l1x,l1y,l1])
    ofuncs2 = set(l2funcs) - set([l2x,l2y,l2])

   
    # List to collect terms in
    terms = []
    if validate: prodterms = []
    # We must expand the expression so that the top-level operation is Add, i.e. it looks like
    # A + B + C + ...
    expr = sympy.expand( expr )
    # What is the top-level operation?
    op = expr.func
    if op is sympy.Add:
        arguments = expr.args # If Add, then we have multiple terms
    else:
        arguments = [expr] # If not Add, then we have a single term
    # Let's factorize each term
    unique_l1s = []
    unique_l2s = []
    
    def homogenize(inexp):
        outexp = inexp.subs([[l1x,Lx],[l2x,Lx],[l1y,Ly],[l2y,Ly],[l1,L],[l2,L]])
        ofuncs = ofuncs1.union(ofuncs2)
        for ofunc in ofuncs:
            nfunc = Symbol(str(ofunc)[:-3])
            outexp = outexp.subs(ofunc,nfunc)
        return outexp

    
    def get_group(inexp):
        if groups is None: return 0
        found = False
        d = Symbol('dummy')
        for i,group in enumerate(groups):
            s = inexp.subs(group,d)
            if not((s/d).has(d)):
                if found:
                    print(s,group)
                    raise ValueError("Groups don't seem to be mutually exclusive.")
                index = i
                found = True
        if not(found):
            raise ValueError("Couldn't associate a group")
        return index


    ogroups = [] if not(groups is None) else None
    ogroup_weights = [] if not(groups is None) else None
    ogroup_symbols = sympy.ones(len(groups),1) if not(groups is None) else None
    for k,arg in enumerate(arguments):
        temp, ll1terms = arg.as_independent(*l1funcs, as_Mul=True)
        loterms, ll2terms = temp.as_independent(*l2funcs, as_Mul=True)

        
        if any([x==0 for x in [ll1terms,ll2terms,loterms]]): continue
        # Group ffts
        if groups is not None:
            gindex = get_group(loterms)
            ogroups.append(gindex)
            fsyms = loterms.free_symbols
            ocoeff = loterms.evalf(subs=dict(zip(fsyms,[1]*len(fsyms))))
            ogroup_weights.append( float(ocoeff) )
            if ogroup_symbols[gindex]==1:
                ogroup_symbols[gindex] = loterms/ocoeff
            else:
                assert ogroup_symbols[gindex]==loterms/ocoeff, "Error validating group membership"
        
        vdict = {}
        vdict['l1'] = ll1terms
        vdict['l2'] = ll2terms
        tdict = {}
        tdict['l1'] = homogenize(vdict['l1'])
        tdict['l2'] = homogenize(vdict['l2'])

        if not(tdict['l1'] in unique_l1s):
            unique_l1s.append(tdict['l1'])
        tdict['l1index'] = unique_l1s.index(tdict['l1'])

        if not(tdict['l2'] in unique_l2s):
            unique_l2s.append(tdict['l2'])
        tdict['l2index'] = unique_l2s.index(tdict['l2'])

        
        
        vdict['other'] = loterms
        tdict['other'] = loterms
        terms.append(tdict)
        # Validate!
        if validate:
            # Check that all the factors of this term do give back the original term
            products = sympy.Mul(vdict['l1'])*sympy.Mul(vdict['l2'])*sympy.Mul(vdict['other'])
            assert sympy.simplify(products-arg)==0, val_fail_message
            prodterms.append(products)
            # Check that the factors don't include symbols they shouldn't
            assert all([not(vdict['l1'].has(x)) for x in l2funcs]), val_fail_message
            assert all([not(vdict['l2'].has(x)) for x in l1funcs]), val_fail_message
            assert all([not(vdict['other'].has(x)) for x in l1funcs]), val_fail_message
            assert all([not(vdict['other'].has(x)) for x in l2funcs]), val_fail_message
    # Check that the sum of products of final form matches original expression
    if validate:
        fexpr = sympy.Add(*prodterms)
        assert sympy.simplify(expr-fexpr)==0, val_fail_message
    return terms,unique_l1s,unique_l2s,ogroups,ogroup_weights,ogroup_symbols



def integrate(shape,wcs,feed_dict,expr,xmask=None,ymask=None,cache=True,validate=True,groups=None,physical_units=True):
    """
    Integrate an arbitrary expression after factorizing it.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays.
    expr: :obj:`sympy.core.symbol.Symbol` 
        A sympy expression containing recognized symbols (see docs)
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    cache: boolean, optional
        Whether to store in memory and reuse repeated terms. Defaults to true.
    validate: boolean,optional
        Whether to check that the final expression and the original agree. Defaults to True.
    groups: list,optional 
        Group all terms such that they have common factors of the provided list of 
        expressions to reduce the number of FFTs.
    physical_units: boolean,optional
        Whether the input is in pixel units or not.

    Returns
    -------

    result : (Ny,Nx) ndarray
        The numerical result of the integration of the expression after factorization.

    """
    # Geometry
    modlmap = enmap.modlmap(shape,wcs)
    lymap,lxmap = enmap.lmap(shape,wcs)

    pixarea = np.prod(enmap.pixshape(shape,wcs))
    feed_dict['L'] = modlmap
    feed_dict['Ly'] = lymap
    feed_dict['Lx'] = lxmap
    shape = shape[-2:]
    ones = enmap.ones(shape,wcs,dtype=np.float32)
    val = 0.
    if xmask is None: 
        warnings.warn("No xmask specified; assuming all ones. This is probably not going to end well.")
        xmask = ones
    if ymask is None: 
        warnings.warn("No xmask specified; assuming all ones. This is probably not going to end well.")
        ymask = ones

    # Expression
    syms = expr.free_symbols
    l1funcs = []
    l2funcs = []
    for sym in syms:
        strsym = str(sym)
        if   strsym[-3:]=="_l1": l1funcs.append(sym)
        elif strsym[-3:]=="_l2": l2funcs.append(sym)

    integrands,ul1s,ul2s, \
        ogroups,ogroup_weights, \
        ogroup_symbols = factorize_2d_convolution_integral(expr,l1funcs=l1funcs,l2funcs=l2funcs,
                                                                         validate=validate,groups=groups)
    def _fft(x): 
        return fft(enmap.enmap(x+0j,wcs))
    def _ifft(x): 
        return ifft(enmap.enmap(x+0j,wcs))

    if cache:
        cached_u1s = []
        cached_u2s = []
        for u1 in ul1s:
            l12d = evaluate(u1,feed_dict)*ones
            cached_u1s.append(_ifft(l12d*xmask))
        for u2 in ul2s:
            l22d = evaluate(u2,feed_dict)*ones
            cached_u2s.append(_ifft(l22d*ymask))


    # For each term, the index of which group it belongs to  

    def get_l1l2(term):
        if cache:
            ifft1 = cached_u1s[term['l1index']]
            ifft2 = cached_u2s[term['l2index']]
        else:
            l12d = evaluate(term['l1'],feed_dict)*ones
            ifft1 = _ifft(l12d*xmask)
            l22d = evaluate(term['l2'],feed_dict)*ones
            ifft2 = _ifft(l22d*ymask)
        return ifft1,ifft2


    if ogroups is None:    
        for i,term in enumerate(integrands):
            ifft1,ifft2 = get_l1l2(term)
            ot2d = evaluate(term['other'],feed_dict)*ones
            ffft = _fft(ifft1*ifft2)
            val += ot2d*ffft
    else:
        vals = np.zeros((len(ogroup_symbols),)+shape,dtype=np.float32)+0j
        for i,term in enumerate(integrands):
            ifft1,ifft2 = get_l1l2(term)
            gindex = ogroups[i]
            vals[gindex,...] += ifft1*ifft2 *ogroup_weights[i]
        for i,group in enumerate(ogroup_symbols):
            ot2d = evaluate(ogroup_symbols[i],feed_dict)*ones            
            ffft = _fft(vals[i,...])
            val += ot2d*ffft
     
    mul = 1 if physical_units else 1./pixarea
    return val * mul






import numpy as np
from scipy.interpolate import interp1d
from pixell import enmap,utils as putils
import sympy
from sympy import Symbol
import warnings
import contextlib
import os,sys
_warn = False

"""
pixell helpers
"""

def mask_kspace(shape,wcs, lxcut = None, lycut = None, lmin = None, lmax = None):
    """Produce a Fourier space mask.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    lxcut : int, optional
        The width of a band in number of Fourier pixels to be masked in the lx direction.
        Default is no masking in this band.
    lycut : int, optional
        The width of a band in number of Fourier pixels to be masked in the ly direction.
        Default is no masking in this band.
    lmin : int, optional
        The radial distance in Fourier space below which all Fourier pixels are masked.
        Default is no masking.
    lmax : int, optional
        The radial distance in Fourier space above which all Fourier pixels are masked.
        Default is no masking.

    Returns
    -------

    output : (Ny,Nx) ndarray
        A 2D array containing the Fourier space mask.

    """
    output = np.ones(shape[-2:], dtype = int)
    if (lmin is not None) or (lmax is not None): modlmap = enmap.modlmap(shape, wcs)
    if (lxcut is not None) or (lycut is not None): ly, lx = enmap.laxes(shape, wcs, oversample=1)
    if lmin is not None:
        output[np.where(modlmap <= lmin)] = 0
    if lmax is not None:
        output[np.where(modlmap >= lmax)] = 0
    if lxcut is not None:
        output[:,np.where(np.abs(lx) < lxcut)] = 0
    if lycut is not None:
        output[np.where(np.abs(ly) < lycut),:] = 0
    return output


def interp(x,y,bounds_error=False,fill_value=0.,**kwargs):
    """Return a function that interpolates (x,y). This wraps around 
    scipy.interpolate.interp1d but by defaulting to zero filling outside bounds.

    Docstring copied from scipy. Interpolate a 1-D function.
    `x` and `y` are arrays of values used to approximate some function f:
    ``y = f(x)``.  This class returns a function whose call method uses
    interpolation to find the value of new points.
    Note that calling `interp1d` with NaNs present in input values results in
    undefined behaviour.
    Parameters
    ----------
    x : (N,) array_like
        A 1-D array of real values.
    y : (...,N,...) array_like
        A N-D array of real values. The length of `y` along the interpolation
        axis must be equal to the length of `x`.
    kind : str or int, optional
        Specifies the kind of interpolation as a string
        ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic',
        'previous', 'next', where 'zero', 'slinear', 'quadratic' and 'cubic'
        refer to a spline interpolation of zeroth, first, second or third
        order; 'previous' and 'next' simply return the previous or next value
        of the point) or as an integer specifying the order of the spline
        interpolator to use.
        Default is 'linear'.
    axis : int, optional
        Specifies the axis of `y` along which to interpolate.
        Interpolation defaults to the last axis of `y`.
    copy : bool, optional
        If True, the class makes internal copies of x and y.
        If False, references to `x` and `y` are used. The default is to copy.
    bounds_error : bool, optional
        If True, a ValueError is raised any time interpolation is attempted on
        a value outside of the range of x (where extrapolation is
        necessary). If False, out of bounds values are assigned `fill_value`.
        False by default.
    fill_value : array-like or (array-like, array_like) or "extrapolate", optional
        - if a ndarray (or float), this value will be used to fill in for
          requested points outside of the data range. If not provided, then
          the default is zero. The array-like must broadcast properly to the
          dimensions of the non-interpolation axes.
        - If a two-element tuple, then the first element is used as a
          fill value for ``x_new < x[0]`` and the second element is used for
          ``x_new > x[-1]``. Anything that is not a 2-element tuple (e.g.,
          list or ndarray, regardless of shape) is taken to be a single
          array-like argument meant to be used for both bounds as
          ``below, above = fill_value, fill_value``.
          .. versionadded:: 0.17.0
        - If "extrapolate", then points outside the data range will be
          extrapolated.
          .. versionadded:: 0.17.0
    assume_sorted : bool, optional
        If False, values of `x` can be in any order and they are sorted first.
        If True, `x` has to be an array of monotonically increasing values.
    Methods
    -------
    __call__
    See Also
    --------
    splrep, splev
        Spline interpolation/smoothing based on FITPACK.
    UnivariateSpline : An object-oriented wrapper of the FITPACK routines.
    interp2d : 2-D interpolation
    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> from scipy import interpolate
    >>> x = np.arange(0, 10)
    >>> y = np.exp(-x/3.0)
    >>> f = interpolate.interp1d(x, y)
    >>> xnew = np.arange(0, 9, 0.1)
    >>> ynew = f(xnew)   # use interpolation function returned by `interp1d`
    >>> plt.plot(x, y, 'o', xnew, ynew, '-')
    >>> plt.show()

    """
    return interp1d(x,y,bounds_error=bounds_error,fill_value=fill_value,**kwargs)

def gauss_beam(ells,fwhm):
    """Return a Gaussian beam transfer function for the given ells.

    Parameters
    ----------

    ells : ndarray
        Any numpy array containing the multipoles at which the beam transfer function
        is requested.
    fwhm : float
        The beam FWHM in arcminutes.

    Returns
    -------

    output : ndarray
        An array of the same shape as ells containing the Gaussian beam transfer function
        for those multipoles.

    """
    tht_fwhm = np.deg2rad(fwhm / 60.)
    return np.exp(-(tht_fwhm**2.)*(ells**2.) / (16.*np.log(2.)))


class bin2D(object):
    def __init__(self, modrmap, bin_edges):
        self.centers = (bin_edges[1:]+bin_edges[:-1])/2.
        self.digitized = np.digitize(np.ndarray.flatten(modrmap), bin_edges,right=True)
        self.bin_edges = bin_edges
    def bin(self,data2d,weights=None):
        if weights is None:
            res = np.bincount(self.digitized,(data2d).reshape(-1))[1:-1]/np.bincount(self.digitized)[1:-1]
        else:
            res = np.bincount(self.digitized,(data2d*weights).reshape(-1))[1:-1]/np.bincount(self.digitized,weights.reshape(-1))[1:-1]
        return self.centers,res

def rect_geometry(width_arcmin=None,width_deg=None,px_res_arcmin=0.5,proj="car",pol=False,height_deg=None,height_arcmin=None,xoffset_degree=0.,yoffset_degree=0.,extra=False,**kwargs):
    """
    Get shape and wcs for a rectangular patch of specified size and coordinate center
    """

    if width_deg is not None:
        width_arcmin = 60.*width_deg
    if height_deg is not None:
        height_arcmin = 60.*height_deg
    
    hwidth = width_arcmin/2.
    if height_arcmin is None:
        vwidth = hwidth
    else:
        vwidth = height_arcmin/2.
    arcmin =  putils.arcmin
    degree =  putils.degree
    pos = [[-vwidth*arcmin+yoffset_degree*degree,-hwidth*arcmin+xoffset_degree*degree],[vwidth*arcmin+yoffset_degree*degree,hwidth*arcmin+xoffset_degree*degree]]
    shape, wcs = enmap.geometry(pos=pos, res=px_res_arcmin*arcmin, proj=proj,**kwargs)
    if pol: shape = (3,)+shape
    if extra:
        modlmap = enmap.modlmap(shape,wcs)
        lmax = modlmap.max()
        ells = np.arange(0,lmax,1.)
        return shape,wcs,modlmap,ells
    else:
        return shape, wcs
    
"""
Sympy helpers
"""

known_zeros = [('E','B'), ('B','E'), ('T','B'), ('B','T')]
def _handle_missing_keys(t1,t2,comp1,comp2):
    if (comp1,comp2) in known_zeros:
        if _warn: warnings.warn("Assuming " + t1 + " is zero. Provide a value for it in feed_dict if not!")
        return 0
    else:
        if "_" in comp1:
            c1s = comp1.split("_")
            c2s = comp2.split("_")
            assert len(c1s)==len(c2s)==2
            return _handle_missing_keys(t1,t2,c1s[1],c2s[1])
        raise KeyError('Neither ',t1, ' nor ', t2, ' were found in feed_dict.')


def get_feed(feed_dict,key):
    # Instead of
    # return feed_dict[k]
    # this allows symmetrized keys.
    
    if "_" not in key: return feed_dict[key]
    
    # Handle l1 and l2 symbols
    if key[-3:]=='_l1' or key[-3:]=='_l2':
        skey = key[:-3]
        suff = key[-3:]
    else:
        skey = key
        suff = ""

    # Split into components
    components = skey.split('_')
    fcomp = components[0]
    if len(components)==3:
        # Non-field-name-case
        comp1 = components[1]
        comp2 = components[2]
    elif len(components)==5:
        # Field-name-case
        comp1 = components[1] + "_" + components[2]
        comp2 = components[3] + "_" + components[4]
    else:
        raise KeyError('This key ',key, ' has an unsupported number of components. Excluding the l1/l2 tag, there should be either three (no field-name case) or five (field-name case).')

    t1 = fcomp + "_" + comp1 + "_" + comp2 + suff
    t2 = fcomp + "_" + comp2 + "_" + comp1 + suff

    try:
        return feed_dict[t1]
    except:
        try:
            return feed_dict[t2]
        except:
            return _handle_missing_keys(t1,t2,comp1,comp2)
    

def evaluate(symbolic_term,feed_dict):
    """
    Convert a symbolic term into a numerical result by using values for
    the symbols from a dictionary.

    symbolic_term: sympy expression
    feed_dict: dictionary mapping names of symbols to numpy arrays
    """
    symbols = list(symbolic_term.free_symbols)
    func_term = sympy.lambdify(symbols,symbolic_term,dummify=False)
    # func_term accepts as keyword arguments strings that are in symbols
    # We need to extract a dict from feed_dict that only has the keywords
    # in symbols
    varstrs = [str(x) for x in symbols]
    edict = {k: get_feed(feed_dict,k) for k in varstrs}
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        evaled = np.nan_to_num(func_term(**edict))
    return evaled

def substitute_trig(l1x,l1y,l2x,l2y,l1,l2):
    # Expand cos(2theta_12) and sin(2theta_12) in terms of l1x,l2x,l1y,l2y,l1,l2
    phi1 = Symbol('phi1')
    phi2 = Symbol('phi2')
    cos2t12 = sympy.cos(2*(phi1-phi2))
    sin2t12 = sympy.sin(2*(phi1-phi2))
    simpcos = sympy.expand_trig(cos2t12)
    simpsin = sympy.expand_trig(sin2t12)
    cos2t12 = sympy.expand(sympy.simplify(simpcos.subs([(sympy.cos(phi1),l1x/l1),(sympy.cos(phi2),l2x/l2),
                            (sympy.sin(phi1),l1y/l1),(sympy.sin(phi2),l2y/l2)])))
    sin2t12 = sympy.expand(sympy.simplify(simpsin.subs([(sympy.cos(phi1),l1x/l1),(sympy.cos(phi2),l2x/l2),
                            (sympy.sin(phi1),l1y/l1),(sympy.sin(phi2),l2y/l2)])))
    return cos2t12,sin2t12




from __future__ import print_function
import numpy as np
from sympy import Symbol,Function
import sympy
from pixell import fft as efft, enmap
import os,sys
from .factorize import Ldl1,Ldl2,l1,l2,cos2t12,sin2t12,l1x,l2x,l1y,l2y,e,L,Lx,Ly,Lxl1,Lxl2,integrate
import warnings

def cross_names(x,y,fn1,fn2,dstr="t"):
    if fn1 is None: e1 = ""
    else:
        assert "_" not in fn1, "Field names cannot have underscores. Sorry! Use a hyphen instead."
        e1 = fn1+"_"
    if fn2 is None: e2 = ""
    else:
        assert "_" not in fn2, "Field names cannot have underscores. Sorry! Use a hyphen instead."
        e2 = fn2+"_"
    return '%sC_%s%s_%s%s' % (dstr,e1,x,e2,y)

def _get_groups(e1,e2=None,noise=True):
    if e2 is None: e2 = e1
    if e1 != e2: return None
    if e1 in ['hu_ok','hdv']:
        if noise:
            return [Lx*Lx,Ly*Ly,Lx*Ly]
        else:
            return [Ly,Lx]
    else:
        return None
    
class HardenedTT(object):
    def __init__(self,shape,wcs,feed_dict,xmask=None,ymask=None,kmask=None,Al=None,hardening='src',estimator='hu_ok'):
        h = hardening
        f_bias,F_bias,_ = get_mc_expressions(hardening,'TT')
        f_phi,F_phi,_ = get_mc_expressions(estimator,'TT')
        f_bh,F_bh,_ = get_mc_expressions(f'{h}-hardened','TT',estimator_to_harden=estimator)
        self.fdict = feed_dict
        # 1 / Response of the biasing agent to the biasing agent
        self.fdict[f'A{h}_{h}_L'] = A_l_custom(shape,wcs,feed_dict,f_bias,F_bias,xmask=xmask,ymask=ymask,groups=None,kmask=kmask)
        # 1 / Response of the biasing agent to CMB lensing
        self.fdict[f'Aphi_{h}_L'] = A_l_custom(shape,wcs,feed_dict,f_phi,F_bias,xmask=xmask,ymask=ymask,groups=None,kmask=kmask)
        self.Al = A_l_custom(shape,wcs,feed_dict,f_bh,F_bh,xmask=xmask,ymask=ymask,groups=None,kmask=kmask) if Al is None else Al
        self.F_bh = F_bh
        self.xmask = xmask
        self.ymask = ymask
        self.kmask = kmask
        self.shape,self.wcs = shape,wcs
    def get_Nl(self):
        return N_l_cross_custom(self.shape,self.wcs,self.fdict,"TT","TT",self.F_bh,self.F_bh,self.F_bh,
                                     xmask=self.xmask,ymask=self.ymask,
                                     Aalpha=self.Al,Abeta=self.Al,groups=None,kmask=self.kmask)
    def reconstruct(self,feed_dict,xname='X_l1',yname='Y_l2',groups=None,physical_units=True):
        uqe = unnormalized_quadratic_estimator_custom(self.shape,self.wcs,feed_dict,
                                                      self.F_bh,xname=xname,yname=yname,
                                                      xmask=self.xmask,ymask=self.ymask,
                                                      groups=groups,physical_units=physical_units)
        return self.Al * uqe * self.kmask
        
        
    

class QE(object):
    """Construct a quadratic estimator such that the normalization is pre-calculated
    and reused.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization 
        calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below).
    estimator: str,optional
        The name of a pre-defined mode-coupling estimator. If this is provided,
        the argument XY is required and the arguments f, F, and groups are ignored.
        e.g. "hu_ok", "hdv" and "shear".
    XY: str,optional
        The XY pair for the requested estimator. Typical examples include "TT" and "EB".
    f: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response. See the Usage guide
        for details. If this is specified, the argument F is required, and the arguments
        estimator, XY and field_names are ignored.
    F: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the estimator filter. See the Usage guide
        for details. 
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    groups: list,optional 
        Group all terms in the normalization such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.


    """
    def __init__(self,shape,wcs,feed_dict,estimator=None,XY=None,
                 f=None,F=None,xmask=None,ymask=None,
                 field_names=None,groups=None,kmask=None):
        """
        """
        if f is not None:
            assert F is not None
            assert estimator is None
            assert XY is None
            self.Al = A_l_custom(shape,wcs,feed_dict,f,F,xmask=xmask,ymask=ymask,groups=groups,kmask=kmask)
            self.F = F
            self.custom = True
        else:
            assert F is None
            self.Al = A_l(shape,wcs,feed_dict,estimator,XY,xmask=xmask,ymask=ymask,field_names=field_names,kmask=kmask)
            self.estimator = estimator
            self.XY = XY
            self.field_names = field_names
            self.custom = False
        self.shape = shape
        self.wcs = wcs
        self.xmask = xmask
        self.ymask = ymask
        kmask = 1 if kmask is None else kmask
        self.kmask = kmask
        
    def reconstruct(self,feed_dict,xname='X_l1',yname='Y_l2',groups=None,physical_units=True):
        """
        Returns a normalized reconstruction corresponding to the initialized
        mode-coupling estimator.

        Parameters
        ----------

        feed_dict: dict
            Mapping from names of custom symbols to numpy arrays used in the reconstruction 
            calculation. When using pre-defined mode-coupling estimators, typical
            keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
            depend on the requested estimator XY (see below). This feed_dict must
            also contain the keys with name xname and yname (see below), which 
            contain the 2D maps X and Y for the data from which the quadratic estimate
            is constructed.
        xname: str,optional
            The name of the key in feed_dict where the X map in the XY quadratic estimator
            is stored. Defaults to X_l1.
        yname: str,optional
            The name of the key in feed_dict where the Y map in the XY quadratic estimator
            is stored. Defaults to Y_l2.
        groups: list,optional 
            Group all terms in the reconstruction calclulation such that they have common factors 
            of the provided list of expressions to reduce the number of FFTs.

        Returns
        -------

        krecon : (Ny,Nx) ndarray
            The normalized Fourier space reconstruction in physical units (not pixel units).
        """
        if self.custom:
            uqe = unnormalized_quadratic_estimator_custom(self.shape,self.wcs,feed_dict,
                                                           self.F,xname=xname,yname=yname,
                                                           xmask=self.xmask,ymask=self.ymask,
                                                           groups=groups,physical_units=physical_units)
        else:
            uqe = unnormalized_quadratic_estimator(self.shape,self.wcs,feed_dict,
                                                   self.estimator,self.XY,
                                                   xname=xname,yname=yname,
                                                   field_names=self.field_names,
                                                   xmask=self.xmask,ymask=self.ymask,physical_units=physical_units)
        return self.Al * uqe * self.kmask

def cross_integral_custom(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,
                           xmask=None,ymask=None,
                           field_names_alpha=None,field_names_beta=None,groups=None,power_name="t"):
    """
    Calculates the integral

    .. math::
        \\int \\frac{d^2 \\vec{l}_1 }{ (2\\pi)^2 }  F_{\\alpha}(\\vec{l}_1,\\vec{l}_2) (F_{\\beta}(\\vec{l}_1,\\vec{l}_2) C^{ac}_{l_1} C^{bd}_{l_2}+ F_{\\beta}(\\vec{l}_2,\\vec{l}_1) C^{ad}_{l_1} C^{bc}_{l_2})

    where
    alpha_XY = "ab"
    beta_XY = "cd"


    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    alpha_XY: str
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    beta_XY: str
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    Falpha: :obj:`sympy.core.symbol.Symbol`
        A sympy expression containing the first alpha estimator filter. See the Usage guide
        for details. 
    Fbeta: :obj:`sympy.core.symbol.Symbol` 
        A sympy expression containing the second beta estimator filter. See the Usage guide
        for details. 
    Fbeta_rev: :obj:`sympy.core.symbol.Symbol` 
        Same as above but with l1 and l2 swapped.
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names_alpha: 2 element list, optional
        Providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    field_names_beta: 2 element list, optional
        As above, but for the second beta estimator.
    groups: list,optional 
        Group all terms in the normalization calclulation such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.

    Returns
    -------

    integral : (Ny,Nx) ndarray
        Returns the integral described above.


    """
    a,b = alpha_XY
    c,d = beta_XY
    fnalpha1,fnalpha2 = field_names_alpha if field_names_alpha is not None else (None,None)
    fnbeta1,fnbeta2 = field_names_beta if field_names_beta is not None else (None,None)
    tCac_l1 = e(cross_names(a,c,fnalpha1,fnbeta1,power_name) + "_l1")
    tCbd_l2 = e(cross_names(b,d,fnalpha2,fnbeta2,power_name) + "_l2")
    tCad_l1 = e(cross_names(a,d,fnalpha1,fnbeta2,power_name) + "_l1")
    tCbc_l2 = e(cross_names(b,c,fnalpha2,fnbeta1,power_name) + "_l2")
    Dexpr1 = tCac_l1*tCbd_l2
    Dexpr2 = tCad_l1*tCbc_l2
    return generic_cross_integral(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,Dexpr1,Dexpr2,
                           xmask=xmask,ymask=ymask,
                           field_names_alpha=field_names_alpha,field_names_beta=field_names_beta,groups=groups)

def generic_cross_integral(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,Dexpr1,Dexpr2,
                           xmask=None,ymask=None,
                           field_names_alpha=None,field_names_beta=None,groups=None):
    """
    Calculates the integral

    .. math::
        \\int \\frac{d^2 \\vec{l}_1 }{ (2\\pi)^2 }  F_{\\alpha}(\\vec{l}_1,\\vec{l}_2) (F_{\\beta}(\\vec{l}_1,\\vec{l}_2) D_1({l_1},{l_2})+ F_{\\beta}(\\vec{l}_2,\\vec{l}_1) D_2({l_1},{l_2}))

    where
    alpha_XY = "ab"
    beta_XY = "cd"


    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    alpha_XY: str
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    beta_XY: str
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    Falpha: :obj:`sympy.core.symbol.Symbol`
        A sympy expression containing the first alpha estimator filter. See the Usage guide
        for details. 
    Fbeta: :obj:`sympy.core.symbol.Symbol` 
        A sympy expression containing the second beta estimator filter. See the Usage guide
        for details. 
    Fbeta_rev: :obj:`sympy.core.symbol.Symbol` 
        Same as above but with l1 and l2 swapped.
    Dexpr1: :obj:`sympy.core.symbol.Symbol`  
        A sympy expression entering in the generic integral.
    Dexpr2: :obj:`sympy.core.symbol.Symbol`  
        A second sympy expression entering in the generic integral.
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names_alpha: 2 element list, optional
        Providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    field_names_beta: 2 element list, optional
        As above, but for the second beta estimator.
    groups: list,optional 
        Group all terms in the normalization calclulation such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.

    Returns
    -------

    integral : (Ny,Nx) ndarray
        Returns the integral described above.


    """
    a,b = alpha_XY
    c,d = beta_XY
    fnalpha1,fnalpha2 = field_names_alpha if field_names_alpha is not None else (None,None)
    fnbeta1,fnbeta2 = field_names_beta if field_names_beta is not None else (None,None)
    expr = Falpha*(Fbeta*Dexpr1+Fbeta_rev*Dexpr2)
    integral = integrate(shape,wcs,feed_dict,expr,xmask=xmask,ymask=ymask,groups=groups,
                         physical_units=False).real * enmap.pixsize(shape,wcs)**0.5 / (np.prod(shape[-2:])**0.5)
    assert np.all(np.isfinite(integral))
    return integral

def N_l_cross_custom(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,
                      xmask=None,ymask=None,
                      field_names_alpha=None,field_names_beta=None,
                      falpha=None,fbeta=None,Aalpha=None,Abeta=None,
                     groups=None,kmask=None,power_name="t"):
    """
    Returns the 2D cross-covariance between two custom mode-coupling estimators. 
    This involves 3 integrals, unless pre-calculated normalizations Al are provided.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    alpha_XY: str
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    beta_XY: str
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    falpha: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response for the first estimator alpha. 
        See the Usage guide
        for details. 
    fbeta: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response for the second estimator beta. 
        See the Usage guide for details. 
    Falpha: :obj:`sympy.core.symbol.Symbol`
        A sympy expression containing the first alpha estimator filter. See the Usage guide
        for details. 
    Fbeta: :obj:`sympy.core.symbol.Symbol` 
        A sympy expression containing the second beta estimator filter. See the Usage guide
        for details. 
    Fbeta_rev: :obj:`sympy.core.symbol.Symbol` 
        Same as above but with l1 and l2 swapped.
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names_alpha: 2 element list, optional
        Providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    field_names_beta: 2 element list, optional
        As above, but for the second beta estimator.
    groups: list,optional 
        Group all terms in the normalization calclulation such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.
    Aalpha: (Ny,Nx) ndarray, optional
        Pre-calculated normalization for the first estimator. This is calculated if
        not provided
    Abeta: (Ny,Nx) ndarray, optional
        Pre-calculated normalization for the second estimator. This is calculated if
        not provided

    Returns
    -------

    Nl : (Ny,Nx) ndarray
        The requested 2D cross-covariance.

    """
    cross_integral = cross_integral_custom(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,
                                           xmask=xmask,ymask=ymask,
                                           field_names_alpha=field_names_alpha,
                                           field_names_beta=field_names_beta,
                                           groups=groups,power_name=power_name)
    return generic_noise_expression(cross_integral,shape,wcs,feed_dict,falpha,fbeta,Falpha,Fbeta, \
                                    xmask,ymask,kmask,Aalpha,Abeta)

def generic_noise_expression(cross_integral,shape,wcs,feed_dict,falpha,fbeta,Falpha,Fbeta,xmask=None,ymask=None,kmask=None,
                             Aalpha=None,Abeta=None):
    """
    returns (1/4) A_alpha * A_beta * cross_integral
    """
    if Aalpha is None: Aalpha = A_l_custom(shape,wcs,feed_dict,falpha,Falpha,xmask=xmask,ymask=ymask,kmask=kmask)
    if Abeta is None: Abeta = A_l_custom(shape,wcs,feed_dict,fbeta,Fbeta,xmask=xmask,ymask=ymask,kmask=kmask)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return 0.25 * Aalpha * Abeta * cross_integral


def RDN0_analytic(shape,wcs,feed_dict,alpha_estimator,alpha_XY,beta_estimator,beta_XY,
                  split_estimator=False,Aalpha=None,Abeta=None,xmask=None,ymask=None,kmask=None,
                  field_names_alpha=None,field_names_beta=None,skip_filter_field_names=False):
    """
    Often lovingly called the `dumb' N0 by ACT lensers, this is the analytic expression
    for the realization-dependependent N0 correction when the noise is isotropic and
    no mask is present.

    feed_dict should have
    dC_T_T, etc. the realized total data power spectrum.
    tC_T_T, etc. should be the total coadd power spectrum used in filters
    uC_T_T, etc. the usual theory spectra for the CMB signal.
    nC_T_T etc. should be the expected total theory power spectrum
    But if split_estimator is true:
    dC_T_T, etc. should be the realized cross-power average.
    nC_T_T etc. should be the expected cross-power, which is usually nC without the instrument noise.
    """
    falpha,Falpha,Falpha_rev = get_mc_expressions(alpha_estimator,alpha_XY,field_names=field_names_alpha if not(skip_filter_field_names) else None)
    fbeta,Fbeta,Fbeta_rev = get_mc_expressions(beta_estimator,beta_XY,field_names=field_names_beta  if not(skip_filter_field_names) else None)
    return RDN0_analytic_generic(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,
                                 falpha=falpha,fbeta=fbeta,Aalpha=Aalpha,Abeta=Abeta,xmask=xmask,ymask=ymask,kmask=kmask,
                                 field_names_alpha=field_names_alpha,field_names_beta=field_names_beta,
                                 split_estimator=split_estimator)

def RDN0_analytic_generic(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,
                          falpha=None,fbeta=None,Aalpha=None,Abeta=None,xmask=None,ymask=None,kmask=None,
                          field_names_alpha=None,field_names_beta=None,split_estimator=False,groups=None):
    """
    Often lovingly called the `dumb' N0 by ACT lensers, this is the analytic expression
    for the realization-dependependent N0 correction when the noise is isotropic and
    no mask is present.

    feed_dict should have
    dC_T_T, etc. the realized total data power spectrum.
    tC_T_T, etc. should be the total coadd power spectrum used in filters
    uC_T_T, etc. the usual theory spectra for the CMB signal.
    nC_T_T etc. should be the expected total theory power spectrum
    But if split_estimator is true:
    dC_T_T, etc. should be the realized cross-power average.
    nC_T_T etc. should be the expected cross-power, which is usually nC without the instrument noise.
    """
    a,b = alpha_XY
    c,d = beta_XY
    fnalpha1,fnalpha2 = field_names_alpha if field_names_alpha is not None else (None,None)
    fnbeta1,fnbeta2 = field_names_beta if field_names_beta is not None else (None,None)
    tCac_l1 = e(cross_names(a,c,fnalpha1,fnbeta1,'n') + "_l1")
    tCbd_l2 = e(cross_names(b,d,fnalpha2,fnbeta2,'n') + "_l2")
    tCad_l1 = e(cross_names(a,d,fnalpha1,fnbeta2,'n') + "_l1")
    tCbc_l2 = e(cross_names(b,c,fnalpha2,fnbeta1,'n') + "_l2")
    dCac_l1 = e(cross_names(a,c,fnalpha1,fnbeta1,'d') + "_l1")
    dCbd_l2 = e(cross_names(b,d,fnalpha2,fnbeta2,'d') + "_l2")
    dCad_l1 = e(cross_names(a,d,fnalpha1,fnbeta2,'d') + "_l1")
    dCbc_l2 = e(cross_names(b,c,fnalpha2,fnbeta1,'d') + "_l2")
    Dexpr1 = dCac_l1*tCbd_l2 + tCac_l1*dCbd_l2 - tCac_l1*tCbd_l2
    Dexpr2 = dCad_l1*tCbc_l2 + tCad_l1*dCbc_l2 - tCad_l1*tCbc_l2
    gint = generic_cross_integral(shape,wcs,feed_dict,alpha_XY,alpha_XY,Falpha,Fbeta,Fbeta_rev,Dexpr1,Dexpr2,
                                  xmask=xmask,ymask=ymask,
                                  field_names_alpha=field_names_alpha,field_names_beta=field_names_beta,groups=groups)
    return generic_noise_expression(gint,shape,wcs,feed_dict,falpha,fbeta,Falpha,Fbeta, \
                                    xmask,ymask,kmask,Aalpha,Abeta)

    
def N_l_cross(shape,wcs,feed_dict,alpha_estimator,alpha_XY,beta_estimator,beta_XY,
              xmask=None,ymask=None,
              Aalpha=None,Abeta=None,field_names_alpha=None,field_names_beta=None,kmask=None,
              skip_filter_field_names=False,power_name="t"):
    """
    Returns the 2D cross-covariance between two pre-defined mode-coupling estimators. 
    This involves 3 integrals, unless pre-calculated normalizations Al are provided.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    alpha_estimator: str
        The name of the first pre-defined mode-coupling estimator. If this is provided,
        the argument XY is required and the arguments f, F, and groups are ignored.
        e.g. "hu_ok", "hdv" and "shear".
    beta_estimator: str,optional
        The name of the second pre-defined mode-coupling estimator. If this is provided,
        the argument XY is required and the arguments f, F, and groups are ignored.
        e.g. "hu_ok", "hdv" and "shear".
    alpha_XY: str,optional
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    beta_XY: str,optional
        The XY pair for the first estimator. Typical examples include "TT" and "EB".
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    Aalpha: (Ny,Nx) ndarray, optional
        Pre-calculated normalization for the first estimator. This is calculated if
        not provided
    Abeta: (Ny,Nx) ndarray, optional
        Pre-calculated normalization for the second estimator. This is calculated if
        not provided
    field_names_alpha: 2 element list, optional
        Providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    field_names_beta: 2 element list, optional
        As above, but for the second beta estimator.


    Returns
    -------

    Nl : (Ny,Nx) ndarray
        The requested 2D cross-covariance.

    """
    falpha,Falpha,Falpha_rev = get_mc_expressions(alpha_estimator,alpha_XY,field_names=field_names_alpha if not(skip_filter_field_names) else None)
    fbeta,Fbeta,Fbeta_rev = get_mc_expressions(beta_estimator,beta_XY,field_names=field_names_beta  if not(skip_filter_field_names) else None)
    return N_l_cross_custom(shape,wcs,feed_dict,alpha_XY,beta_XY,Falpha,Fbeta,Fbeta_rev,
                            xmask=xmask,ymask=ymask,
                            field_names_alpha=field_names_alpha,field_names_beta=field_names_beta,
                            falpha=falpha,fbeta=fbeta,Aalpha=Aalpha,Abeta=Abeta,
                            groups=_get_groups(alpha_estimator,beta_estimator),kmask=kmask,
                            power_name=power_name)

def N_l(shape,wcs,feed_dict,estimator,XY,
        xmask=None,ymask=None,
        Al=None,field_names=None,kmask=None,power_name="t"):
    """
    Returns the 2D noise corresponding to a pre-defined mode-coupling estimator
    NOT assuming that it is optimal. This involves 2 integrals, unless a pre-calculated
    normalization Al is provided, in which case only 1 integral needs to be evaluated.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    estimator: str
        The name of a pre-defined mode-coupling estimator. 
        e.g. "hu_ok", "hdv" and "shear".
    XY: str
        The XY pair for the requested estimator. Typical examples include "TT" and "EB".
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    Al: (Ny,Nx) ndarray, optional
        Pre-calculated normalization for the estimator. Reduces the number of integrals 
        calculated to 1 if provided, else calculates 2 integrals.
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.

    Returns
    -------

    Nl : (Ny,Nx) ndarray
        The 2D noise for the estimator.

    """
    falpha,Falpha,Falpha_rev = get_mc_expressions(estimator,XY,field_names=field_names)
    return N_l_cross_custom(shape,wcs,feed_dict,XY,XY,Falpha,Falpha,Falpha_rev,
                            xmask=xmask,ymask=ymask,
                            field_names_alpha=field_names,field_names_beta=field_names,
                            falpha=falpha,fbeta=falpha,Aalpha=Al,Abeta=Al,
                            groups=_get_groups(estimator),kmask=kmask,power_name=power_name)

    
def A_l_custom(shape,wcs,feed_dict,f,F,xmask=None,ymask=None,groups=None,kmask=None):
    """
    Returns the 2D normalization corresponding to a custom
    mode-coupling estimator.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    f: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response. See the Usage guide
        for details. If this is specified, the argument F is required, and the arguments
        estimator, XY and field_names are ignored.
    F: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the estimator filter. See the Usage guide
        for details. 
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    xname: str,optional
        The name of the key in feed_dict where the X map in the XY quadratic estimator
        is stored. Defaults to X_l1.
    yname: str,optional
        The name of the key in feed_dict where the Y map in the XY quadratic estimator
        is stored. Defaults to Y_l2.
    groups: list,optional 
        Group all terms in the normalization calclulation such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.

    Returns
    -------

    Al : (Ny,Nx) ndarray
        The 2D normalization for the estimator.


    """
    integral = integrate(shape,wcs,feed_dict,f*F/L/L,
                         xmask=xmask,ymask=ymask,groups=groups,
                         physical_units=False).real * enmap.pixsize(shape,wcs)**0.5 / (np.prod(shape[-2:])**0.5)
    assert np.all(np.isfinite(integral))
    kmask = 1 if kmask is None else kmask
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return np.nan_to_num(1/integral)*kmask
    
def A_l(shape,wcs,feed_dict,estimator,XY,xmask=None,ymask=None,field_names=None,kmask=None):
    """
    Returns the normalization corresponding to a pre-defined mode-coupling estimator.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    estimator: str
        The name of a pre-defined mode-coupling estimator. 
        e.g. "hu_ok", "hdv" and "shear".
    XY: str
        The XY pair for the requested estimator. Typical examples include "TT" and "EB".
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.

    Returns
    -------

    Al : (Ny,Nx) ndarray
        The 2D normalization for the estimator.

    """
    
    f,F,Fr = get_mc_expressions(estimator,XY,field_names=field_names)
    return A_l_custom(shape,wcs,feed_dict,f,F,xmask=xmask,ymask=ymask,groups=_get_groups(estimator),kmask=kmask)

def N_l_from_A_l_optimal(shape,wcs,Al):
    modlmap = enmap.modlmap(shape,wcs)
    return Al * modlmap*(modlmap+1.)/4.    

def N_l_optimal(shape,wcs,feed_dict,estimator,XY,xmask=None,ymask=None,field_names=None,kmask=None):
    """
    Returns the 2D noise corresponding to a pre-defined mode-coupling estimator
    but assuming that it is optimal, i.e.
    Nl = A_L L^2 / 4

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    estimator: str
        The name of a pre-defined mode-coupling estimator. 
        e.g. "hu_ok", "hdv" and "shear".
    XY: str
        The XY pair for the requested estimator. Typical examples include "TT" and "EB".
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.

    Returns
    -------

    Nl : (Ny,Nx) ndarray
        The 2D noise for the estimator.

    """
    
    Al = A_l(shape,wcs,feed_dict,estimator,XY,xmask,ymask,field_names=field_names,kmask=kmask)
    modlmap = enmap.modlmap(shape,wcs)
    return N_l_from_A_l_optimal(shape,wcs,Al)

def N_l_optimal_custom(shape,wcs,feed_dict,f,F,xmask=None,ymask=None,groups=None,kmask=None):
    """
    Returns the 2D noise corresponding to a custom
    mode-coupling estimator but assuming that it is optimal, i.e.
    Nl = A_L L^2 / 4

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    f: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response. See the Usage guide
        for details. If this is specified, the argument F is required, and the arguments
        estimator, XY and field_names are ignored.
    F: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the estimator filter. See the Usage guide
        for details. 
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    xname: str,optional
        The name of the key in feed_dict where the X map in the XY quadratic estimator
        is stored. Defaults to X_l1.
    yname: str,optional
        The name of the key in feed_dict where the Y map in the XY quadratic estimator
        is stored. Defaults to Y_l2.
    groups: list,optional 
        Group all terms in the normalization calclulation such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.

    Returns
    -------

    Nl : (Ny,Nx) ndarray
        The 2D noise for the estimator.


    """
    
    Al = A_l_custom(shape,wcs,feed_dict,f,F,xmask,ymask,groups=groups,kmask=kmask)
    return N_l_from_A_l_optimal(shape,wcs,Al)

def unnormalized_quadratic_estimator_custom(shape,wcs,feed_dict,F,xname='X_l1',yname='Y_l2',
                                            xmask=None,ymask=None,groups=None,physical_units=True):
    """
    Returns a normalized reconstruction corresponding to a custom
    mode-coupling estimator.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    F: :obj:`sympy.core.symbol.Symbol`
        A sympy expression containing the estimator filter. See the Usage guide
        for details. 
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    xname: str,optional
        The name of the key in feed_dict where the X map in the XY quadratic estimator
        is stored. Defaults to X_l1.
    yname: str,optional
        The name of the key in feed_dict where the Y map in the XY quadratic estimator
        is stored. Defaults to Y_l2.
    groups: list,optional 
        Group all terms in the reconstruction calclulation such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.

    Returns
    -------

    krecon : (Ny,Nx) ndarray
        The normalized Fourier space reconstruction in physical units (not pixel units).

    See Also
    --------
    QE
        A class with which the slow normalization can be pre-calculated and repeated estimation
        can be performed on similar datasets.

    """
    
    res = integrate(shape,wcs,feed_dict,e(xname)*e(yname)*F/2,xmask=xmask,ymask=ymask,groups=groups,physical_units=physical_units)
    assert np.all(np.isfinite(res))
    return res

def unnormalized_quadratic_estimator(shape,wcs,feed_dict,estimator,XY,
                                     xname='X_l1',yname='Y_l2',field_names=None,xmask=None,ymask=None,physical_units=True):
    """
    Returns a normalized reconstruction corresponding to specified pre-defined
    mode-coupling estimator.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    estimator: str,optional
        The name of a pre-defined mode-coupling estimator. If this is provided,
        the argument XY is required and the arguments f, F, and groups are ignored.
        e.g. "hu_ok", "hdv" and "shear".
    XY: str,optional
        The XY pair for the requested estimator. Typical examples include "TT" and "EB".
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    xname: str,optional
        The name of the key in feed_dict where the X map in the XY quadratic estimator
        is stored. Defaults to X_l1.
    yname: str,optional
        The name of the key in feed_dict where the Y map in the XY quadratic estimator
        is stored. Defaults to Y_l2.

    Returns
    -------

    krecon : (Ny,Nx) ndarray
        The normalized Fourier space reconstruction in physical units (not pixel units).

    See Also
    --------
    reconstruct
        Get the properly normalized quadratic estimator reconstruction.
    QE
        A class with which the slow normalization can be pre-calculated and repeated estimation
        can be performed on similar datasets.

    """

    
    f,F,Fr = get_mc_expressions(estimator,XY,field_names=field_names)
    return unnormalized_quadratic_estimator_custom(shape,wcs,feed_dict,F,xname=xname,yname=yname,xmask=xmask,ymask=ymask,groups=_get_groups(estimator,noise=False),physical_units=physical_units)


def reconstruct(shape,wcs,feed_dict,estimator=None,XY=None,
                f=None,F=None,xmask=None,ymask=None,
                field_names=None,norm_groups=None,est_groups=None,
                xname='X_l1',yname='Y_l2',kmask=None,physical_units=True):


    """
    Returns a normalized reconstruction corresponding to specified
    mode-coupling estimator.

    Parameters
    ----------

    shape : tuple
        The shape of the array for the geometry of the footprint. Typically 
        (...,Ny,Nx) for Ny pixels in the y-direction and Nx in the x-direction.
    wcs : :obj:`astropy.wcs.wcs.WCS`
        The wcs object completing the specification of the geometry of the footprint.
    feed_dict: dict
        Mapping from names of custom symbols to numpy arrays used in the normalization and
        reconstruction  calculation. When using pre-defined mode-coupling estimators, typical
        keys that must be present are 'uC_X_Y' and 'tC_X_Y', where X and Y
        depend on the requested estimator XY (see below). This feed_dict must
        also contain the keys with name xname and yname (see below), which 
        contain the 2D maps X and Y for the data from which the quadratic estimate
        is constructed.
    estimator: str,optional
        The name of a pre-defined mode-coupling estimator. If this is provided,
        the argument XY is required and the arguments f, F, and groups are ignored.
        e.g. "hu_ok", "hdv" and "shear".
    XY: str,optional
        The XY pair for the requested estimator. Typical examples include "TT" and "EB".
    f: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response. See the Usage guide
        for details. If this is specified, the argument F is required, and the arguments
        estimator, XY and field_names are ignored.
    F: :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the estimator filter. See the Usage guide
        for details. 
    xmask: (Ny,Nx) ndarray,optional
        Fourier space 2D mask for the l1 part of the integral. Defaults to ones.
    ymask:  (Ny,Nx) ndarray, optional
        Fourier space 2D mask for the l2 part of the integral. Defaults to ones.
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more custom noise correlations.
    norm_groups: list,optional 
        Group all terms in the normalization such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.
    xname: str,optional
        The name of the key in feed_dict where the X map in the XY quadratic estimator
        is stored. Defaults to X_l1.
    yname: str,optional
        The name of the key in feed_dict where the Y map in the XY quadratic estimator
        is stored. Defaults to Y_l2.
    est_groups: list,optional 
        Group all terms in the reconstruction calclulation such that they have common factors 
        of the provided list of expressions to reduce the number of FFTs.

    Returns
    -------

    krecon : (Ny,Nx) ndarray
        The normalized Fourier space reconstruction in physical units (not pixel units).

    See Also
    --------
    QE
        A class with which the slow normalization can be pre-calculated and repeated estimation
        can be performed on similar datasets.

    """
    
    qe = QE(shape,wcs,feed_dict,estimator=estimator,XY=XY,
            f=f,F=F,xmask=xmask,ymask=ymask,
            field_names=field_names,groups=norm_groups,kmask=kmask)
    return qe.reconstruct(feed_dict,xname=xname,yname=yname,groups=est_groups,physical_units=physical_units)
    



def u1(ab):
    a,b = ab
    return e('uC_%s_%s_l1' % (a,b))
def u2(ab):
    a,b = ab
    return e('uC_%s_%s_l2' % (a,b))
def du1(ab):
    a,b = ab
    return e('duC_%s_%s_l1' % (a,b))
def du2(ab):
    a,b = ab
    return e('duC_%s_%s_l2' % (a,b))


def lensing_response_f(XY,rev=False,curl=False):

    """
    Returns the mode-coupling response f(l1,l2) for CMB lensing.

    Parameters
    ----------

    XY: str
        The XY pair for the requested estimator. This must belong to one
        of TT, EE, TE, ET, EB or TB.
    rev: boolean, optional
        Whether to swap l1 and l2. Defaults to False.

    Returns
    -------

    f : :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response. See the Usage guide
        for details. 

    """
    
    cLdl1 = Lxl1 if curl else Ldl1
    cLdl2 = Lxl2 if curl else Ldl2
    
    iLdl1 = cLdl2 if rev else cLdl1
    iLdl2 = cLdl1 if rev else cLdl2
    def iu1(ab): return u2(ab) if rev else u1(ab)
    def iu2(ab): return u1(ab) if rev else u2(ab)
    if XY=='TT':
        f = iLdl1*iu1('TT')+iLdl2*iu2('TT')
    elif XY=='TE':
        f = iLdl1*cos2t12*iu1('TE')+iLdl2*iu2('TE')
    elif XY=='ET':
        f = iLdl2*iu2('TE')*cos2t12+iLdl1*iu1('TE')
    elif XY=='TB':
        f = iu1('TE')*sin2t12*iLdl1
    elif XY=='EE':
        f = (iLdl1*iu1('EE')+iLdl2*iu2('EE'))*cos2t12
    elif XY=='EB':
        f = iLdl1*iu1('EE')*sin2t12
    else:
        print(XY)
        raise ValueError
    return f


def rotation_response_f(XY,rev=False):

    """
    Returns the mode-coupling response f(l1,l2) for CMB rotation.

    Parameters
    ----------

    XY: str
        The XY pair for the requested estimator. This must belong to one
        of EE, TE, ET, EB or TB.
    rev: boolean, optional
        Whether to swap l1 and l2. Defaults to False.

    Returns
    -------

    f : :obj:`sympy.core.symbol.Symbol`
        A sympy expression containing the mode-coupling response. See the Usage guide
        for details.

    """

    def iu1(ab): return u2(ab) if rev else u1(ab)
    def iu2(ab): return u1(ab) if rev else u2(ab)

    if XY=='TE':
        f = 2*iu1('TE')*sin2t12
    elif XY=='TB':
        f = 2*iu1('TE')*cos2t12
    elif XY=='EE':
        f = 2*(iu1('EE')-iu2('EE'))*sin2t12
    elif XY=='EB':
        f = 2*(iu1('EE')-iu2('BB'))*cos2t12
    elif XY=='BB':
        f = (iu1('BB')+iu2('BB'))*sin2t12
    else:
        print(XY)
        raise ValueError
    return f

def get_mc_expressions(estimator,XY,field_names=None,estimator_to_harden='hu_ok'):
    """
    Pre-defined mode coupling expressions.
    Returns f(l1,l2), F(l1,l2), F(l2,l1).
    If a list field_names is provided containing two strings,
    then "total power" spectra are customized to potentially
    be different and feed_dict will need to have more values.

    Parameters
    ----------

    estimator: str
        The name of a pre-defined mode-coupling estimator. If this is provided,
        the argument XY is required and the arguments f, F, and groups are ignored.
        e.g. "hu_ok", "hdv" and "shear".
    XY: str
        The XY pair for the requested estimator. Typical examples include "TT" and "EB".
    field_names: 2 element list, optional
        When a pre-defined mode-coupling estimator is used, providing a list field_names
        modifies the total power spectra variable names that feed_dict expects. 
        Typically, names like "tC_T_T" and "tC_T_E" are expected. But if field_names
        is ["E1","E2"] for example, variable names like ``tC_E1_T_E1_T``, ``tC_E2_T_E2_T``,
        ``tC_E1_T_E2_T``, ``tC_E2_T_E1_T`` are expected to be present in feed_dict. This
        allows for more general noise correlations.

    Returns
    -------

    f : :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the mode-coupling response. See the Usage guide
        for details. 
    F : :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the estimator filter. 
    Fr : :obj:`sympy.core.symbol.Symbol` , optional
        A sympy expression containing the estimator filter but with l1 and l2 swapped. 


    """
    
    estimator = estimator.lower()
    X,Y = XY
    XX = X+X
    YY = Y+Y
    f1,f2 = field_names if field_names is not None else (None,None)

    # NOTE: previously, this used cross(f1,f2) for t1 and t2
    # I've switched it to t1 = cross(f1,f1) and t2 = cross(f2,f2)
    # to allow for different filters for the two fields.
    def t1(ab):
        a,b = ab
        return e(cross_names(a,b,f1,f1)+"_l1")
    def t2(ab):
        a,b = ab
        return e(cross_names(a,b,f2,f2)+"_l2")
    
    hus = ['hu_ok_curl','hdv_curl','hu_ok','hdv']
    curls = ['hu_ok_curl','hdv_curl']
    curl = estimator in curls
    cLdl1 = Lxl1 if curl else Ldl1
    cLdl2 = Lxl2 if curl else Ldl2
    if estimator in hus: # Hu, Okamoto 2001 # Hu, DeDeo, Vale 2007
        f = lensing_response_f(XY,rev=False,curl=curl)
        if estimator in ['hu_ok','hu_ok_curl']:
            fr = lensing_response_f(XY,rev=True,curl=curl)
            if XY in ['TT','EE']:
                F = f / 2 / t1(XY) / t2(XY)
                Fr = fr / 2 / t2(XY) / t1(XY)
            elif XY in ['TB','EB']:
                F = f  / t1(XX) / t2(YY)
                Fr = fr  / t2(XX) / t1(YY)
            elif XY=='TE':
                # this filter is not separable
                #(tCl1['EE']*tCl2['TT']*f - tCl1['TE']*tCl2['TE']*frev)/(tCl1['TT']*tCl2['EE']*tCl1['EE']*tCl2['TT']-(tCl1['TE']*tCl2['TE'])**2.)
                # this approximation is
                F = (t1('EE')*t2('TT')*f - t1('TE')*t2('TE')*fr)/(t1('TT')*t2('EE')*t1('EE')*t2('TT'))
                Fr = (t2('EE')*t1('TT')*fr - t2('TE')*t1('TE')*f)/(t2('TT')*t1('EE')*t2('EE')*t1('TT'))
        elif estimator in ['hdv','hdv_curl']:
            Fp = cLdl1/t1(X+X)/t2(Y+Y)
            Fpr = cLdl2/t2(X+X)/t1(Y+Y)
            if Y=='T':
                F = Fp * u1(X+Y)
                Fr = Fpr * u2(X+Y)
            if Y=='E':
                F = Fp * u1(X+Y) * cos2t12
                Fr = Fpr * u2(X+Y) * cos2t12
            if Y=='B':
                F = Fp * u1(X+'E') * sin2t12
                Fr = Fpr * u2(X+'E') * sin2t12

    elif estimator=='shear': # Schaan, Ferraro 2018
        assert XY=="TT", "Shear estimator only implemented for TT."
        f = cLdl2*u2('TT')
        fr = cLdl1*u1('TT')
        cos2theta = ((2*(cLdl1)**2)/L**2/l1**2) - 1
        cos2theta_rev = ((2*(cLdl2)**2)/L**2/l2**2) - 1
        F = cos2theta * u1('TT') * du1('TT')/2/t1('TT')/t1('TT')
        Fr = cos2theta_rev * u2('TT') * du2('TT')/2/t2('TT')/t2('TT')

    elif estimator=='src':
        f = e('pc_T_T_l1')*e('pc_T_T_l2')
        F = f / t1(XY) / t2(XY) / 2
        fr = f
        Fr = F
    elif estimator=='src-hardened':
        """ Osborne et. al. point source hardening
        This gives you the MC expressions for the source
        hardened lensing estimator.
        You have to provide the following during the
        calculation apart from the usual spectra: 
        (1) pc_T_T, the Fourier space profile
        of the source, which is 1 for a point source.
        (2) Asrc_src_L : 1 / the source estimator response to sources
        (3) Aphi_src_L : 1 / the lens estimator response to sources
        """
        assert XY=="TT", "BH only implemented for TT."
        f_phi,F_phi,_ = get_mc_expressions(estimator_to_harden,XY,field_names=field_names)
        f_src,_,_ = get_mc_expressions('src',XY,field_names=field_names)
        A_src_src = e('Asrc_src_L')
        A_phi_src = e('Aphi_src_L')
        f = f_phi - A_src_src / A_phi_src * f_src
        F = f / t1(XY) / t2(XY) / 2
        fr = f
        Fr = F
    elif estimator=='mask': # Namikawa et. al. mask bias hardening
        f = - t1('TT') - t2('TT')
        fr = f
        F = f / t1(XY) / t2(XY) / 2
        fr = f
        Fr = F
    elif estimator=='mask-hardened':
        """ Namikawa et. al. mask hardening
        This gives you the MC expressions for the mask
        hardened lensing estimator.
        You have to provide the following during the
        calculation apart from the usual spectra: 
        (1) Amsk_msk_L : 1 / the mask estimator response to masks
        (2) Aphi_msk_L : 1 / the lens estimator response to masks
        """
        assert XY=="TT", "BH only implemented for TT."
        f_phi,F_phi,_ = get_mc_expressions(estimator_to_harden,XY,field_names=field_names)
        f_msk,_,_ = get_mc_expressions('mask',XY,field_names=field_names)
        A_msk_msk = e('Amask_mask_L')
        A_phi_msk = e('Aphi_mask_L')
        f = f_phi - A_msk_msk / A_phi_msk * f_msk
        F = f / t1(XY) / t2(XY) / 2
        fr = f
        Fr = F
    elif estimator=='rot':  # Yadav et. al. 2009
        f = rotation_response_f(XY,rev=False)
        fr = rotation_response_f(XY,rev=True)
        if XY in ['EE','BB']:
            F = f / 2 / t1(XY) / t2(XY)
            Fr = fr / 2 / t2(XY) / t1(XY)
        elif XY in ['TB','EB']:
            F = f / t1(XX) / t2(YY)
            Fr = fr / t2(XX) / t1(YY)
        elif XY=='TE':
            F = (t1('EE')*t2('TT')*f - t1('TE')*t2('TE')*fr)/(t1('TT')*t2('EE')*t1('EE')*t2('TT'))
            Fr = (t2('EE')*t1('TT')*fr - t2('TE')*t1('TE')*f)/(t2('TT')*t1('EE')*t2('EE')*t1('TT'))

    return f,F,Fr



"""Top-level package for symlens."""

__author__ = """Mathew Madhavacheril"""
__email__ = 'mathewsyriac@gmail.com'
__version__ = '0.3.3'


try:
    from .factorize import *
    from .qe import *
    from .utils import *
except: pass




import symlens as s
from pixell import enmap, utils as putils, powspec
import os,sys
from scipy.interpolate import interp1d
import numpy as np
import pytest


"""
Only tests marked @pytest.mark.webtest
will be run on Travis
"""

@pytest.mark.webtest
def test_trivial():
    assert 1==1


# The following tests need orphics, and so are not meant to be run on Travis
def test_hdv_huok_planck():
    from orphics import lensing,io,cosmology,maps

    shape,wcs = enmap.geometry(shape=(512,512),res=2.0*putils.arcmin,pos=(0,0))
    modlmap = enmap.modlmap(shape,wcs)
    theory = cosmology.default_theory()
    ells = np.arange(0,3000,1)
    ctt = theory.lCl('TT',ells)
    # ps,_ = powspec.read_camb_scalar("tests/Aug6_highAcc_CDM_scalCls.dat")
    # ells = range(ps.shape[-1])

    ## Build HuOk TT estimator
    f = s.Ldl1 * s.e('uC_T_T_l1') + s.Ldl2 * s.e('uC_T_T_l2')
    F = f / 2 / s.e('tC_T_T_l1') / s.e('tC_T_T_l2')
    expr1 = f * F
    feed_dict = {}
    feed_dict['uC_T_T'] = s.interp(ells,ctt)(modlmap)
    feed_dict['tC_T_T'] = s.interp(ells,ctt)(modlmap)+(33.*np.pi/180./60.)**2./s.gauss_beam(modlmap,7.0)**2.
    tellmin = 10 ; tellmax = 3000
    xmask = s.mask_kspace(shape,wcs,lmin=tellmin,lmax=tellmax)
    integral = s.integrate(shape,wcs,feed_dict,expr1,xmask=xmask,ymask=xmask).real
    Nl = modlmap**4./integral/4.
    bin_edges = np.arange(10,3000,40)
    binner = s.bin2D(modlmap,bin_edges)
    cents,nl1d = binner.bin(Nl)
    

    
    ## Build HDV TT estimator
    F = s.Ldl1 * s.e('uC_T_T_l1') / s.e('tC_T_T_l1') / s.e('tC_T_T_l2')
    expr1 = f*F
    integral = s.integrate(shape,wcs,feed_dict,expr1,xmask=xmask,ymask=xmask).real
    Nl = modlmap**4./integral/4.

    cents,nl1d2 = binner.bin(Nl)

    cents,nl1d3 = binner.bin(s.N_l_cross(shape,wcs,feed_dict,"hu_ok","TT","hu_ok","TT",
                                         xmask=xmask,ymask=xmask))
    cents,nl1d4 = binner.bin(s.N_l_cross(shape,wcs,feed_dict,"hdv","TT","hdv","TT",
                                       xmask=xmask,ymask=xmask))
    cents,nl1d5 = binner.bin(s.N_l(shape,wcs,feed_dict,"hu_ok","TT",xmask=xmask,ymask=xmask))
    cents,nl1d6 = binner.bin(s.N_l(shape,wcs,feed_dict,"hdv","TT",xmask=xmask,ymask=xmask))

    clkk = theory.gCl('kk',ells)
    pl = io.Plotter(xyscale='linlog',xlabel='L',ylabel='C')
    pl.add(cents,nl1d)
    pl.add(cents,nl1d2)
    # pl.add(cents,nl1d3)
    pl.add(cents,nl1d4)
    pl.add(cents,nl1d5)
    pl.add(cents,nl1d6)
    pl.add(ells,clkk)
    pl.done("plcomp.png")
    

def test_lens_recon():
    from orphics import lensing,io,cosmology,maps
    from enlib import bench

    deg = 10.
    px = 2.0
    tellmin = 100
    tellmax = 3000
    kellmin = 40
    kellmax = 3000
    grad_cut = None
    bin_width = 80
    beam_arcmin = 0.01
    noise_uk_arcmin = 0.01
    
    theory = cosmology.default_theory(lpad=30000)
    shape,wcs = s.rect_geometry(width_deg=deg,px_res_arcmin=px)
    flsims = lensing.FlatLensingSims(shape,wcs,theory,beam_arcmin,noise_uk_arcmin)
    kbeam = flsims.kbeam
    modlmap = enmap.modlmap(shape,wcs)
    fc = maps.FourierCalc(shape,wcs)
    n2d = (noise_uk_arcmin*np.pi/180./60.)**2./flsims.kbeam**2.
    tmask = s.mask_kspace(shape,wcs,lmin=tellmin,lmax=tellmax)
    kmask = s.mask_kspace(shape,wcs,lmin=kellmin,lmax=kellmax)
    with bench.show("orphics init"):
        qest = lensing.qest(shape,wcs,theory,
                            noise2d=n2d,kmask=tmask,kmask_K=kmask,
                            pol=False,grad_cut=grad_cut,
                            unlensed_equals_lensed=True,bigell=30000)
    bin_edges = np.arange(kellmin,kellmax,bin_width)
    binner = s.bin2D(modlmap,bin_edges)
    i = 0
    unlensed,kappa,lensed,beamed,noise_map,observed = flsims.get_sim(seed_cmb=(i,1),
                                                                     seed_kappa=(i,2),
                                                                     seed_noise=(i,3),
                                                                     lens_order=5,
                                                                     return_intermediate=True)

    kmap = enmap.fft(observed,normalize="phys")
    # _,kmap,_ = fc.power2d(observed)
    with bench.show("orphics"):
        kkappa = qest.kappa_from_map("TT",kmap/kbeam,alreadyFTed=True,returnFt=True)    
    pir2d,kinput = fc.f1power(kappa,kkappa)
    pii2d = fc.f2power(kinput,kinput)
    prr2d = fc.f2power(kkappa,kkappa)
    cents,pir1d = binner.bin(pir2d)
    cents,pii1d = binner.bin(pii2d)
    cents,prr1d = binner.bin(prr2d)

    feed_dict = {}
    cltt = theory.lCl('TT',modlmap)
    feed_dict['uC_T_T'] = theory.lCl('TT',modlmap)
    feed_dict['tC_T_T'] = cltt+n2d
    feed_dict['X'] = kmap/kbeam
    feed_dict['Y'] = kmap/kbeam

    with bench.show("symlens init"):
        Al = s.A_l(shape,wcs,feed_dict,"hdv","TT",xmask=tmask,ymask=tmask)
    Nl = s.N_l_from_A_l_optimal(shape,wcs,Al)
    with bench.show("symlens"):
        ukappa = s.unnormalized_quadratic_estimator(shape,wcs,feed_dict,"hdv","TT",xmask=tmask,ymask=tmask)
    nkappa = Al * ukappa

    pir2d2 = fc.f2power(nkappa,kinput)
    cents,pir1d2 = binner.bin(pir2d2)


    cents,Nlkk = binner.bin(qest.N.Nlkk['TT'])
    cents,Nlkk2 = binner.bin(Nl)

    pl = io.Plotter(xyscale='linlog',xlabel='L',ylabel='C')
    pl.add(cents,pii1d,color='k',lw=3)
    pl.add(cents,pir1d,label='orphics')
    pl.add(cents,pir1d2,label='hdv symlens')
    pl.add(cents,Nlkk,ls="--",label='orphics')
    pl.add(cents,Nlkk2,ls="-.",label='symlens')
    pl.done("ncomp.png")

def test_shear():
    from orphics import lensing,io,cosmology,maps

    deg = 20.
    px = 2.0
    tellmin = 30
    tellmax = 3500
    kellmin = 10
    kellmax = 3000
    bin_width = 20
    beam_arcmin = 1.4
    noise_uk_arcmin = 7.0
    
    theory = cosmology.default_theory(lpad=30000)
    shape,wcs = s.rect_geometry(width_deg=deg,px_res_arcmin=px)
    flsims = lensing.FlatLensingSims(shape,wcs,theory,beam_arcmin,noise_uk_arcmin)
    kbeam = flsims.kbeam
    modlmap = enmap.modlmap(shape,wcs)
    fc = maps.FourierCalc(shape,wcs)
    n2d = (noise_uk_arcmin*np.pi/180./60.)**2./flsims.kbeam**2.
    tmask = s.mask_kspace(shape,wcs,lmin=tellmin,lmax=tellmax)
    kmask = s.mask_kspace(shape,wcs,lmin=kellmin,lmax=kellmax)
    bin_edges = np.arange(kellmin,kellmax,bin_width)
    binner = s.bin2D(modlmap,bin_edges)
    i = 0
    unlensed,kappa,lensed,beamed,noise_map,observed = flsims.get_sim(seed_cmb=(i,1),
                                                                     seed_kappa=(i,2),
                                                                     seed_noise=(i,3),
                                                                     lens_order=5,
                                                                     return_intermediate=True)
    _,kmap,_ = fc.power2d(observed)
    pii2d,kinput,_ = fc.power2d(kappa)

    feed_dict = {}
    cltt = theory.lCl('TT',modlmap)
    feed_dict['uC_T_T'] = theory.lCl('TT',modlmap)
    feed_dict['tC_T_T'] = (cltt+n2d)
    feed_dict['X'] = kmap/kbeam
    feed_dict['Y'] = kmap/kbeam

    ells = np.arange(0,10000,1)
    ucltt = theory.lCl('TT',ells)
    feed_dict['duC_T_T'] = s.interp(ells,np.gradient(np.log(ucltt),np.log(ells)))(modlmap)
    sAl = s.A_l(shape,wcs,feed_dict,"shear","TT",xmask=tmask,ymask=tmask)
    sNl = s.N_l(shape,wcs,feed_dict,"shear","TT",
              xmask=tmask,ymask=tmask,
              Al=sAl)
    sukappa = s.unnormalized_quadratic_estimator(shape,wcs,feed_dict,"shear","TT",xmask=tmask,ymask=tmask)
    snkappa = sAl * sukappa
    

    pir2d3 = fc.f2power(snkappa,kinput)
    cents,pir1d3 = binner.bin(pir2d3)
    cents,pii1d = binner.bin(pii2d)
    cents,prr1d = binner.bin(fc.f2power(snkappa,snkappa))
    
    cents,Nlkk3 = binner.bin(sNl)

    pl = io.Plotter(xyscale='loglog',xlabel='L',ylabel='C')
    pl.add(ells,theory.gCl('kk',ells))
    pl.add(cents,pii1d,color='k',lw=3)
    pl.add(cents,pir1d3,label='shear')
    pl.add(cents,prr1d)
    pl.add(cents,Nlkk3,ls=":")
    pl._ax.set_xlim(10,3500)
    pl.done("ncomp.png")
    
def test_pol():
    from orphics import lensing,io,cosmology,maps

    est = "hu_ok"
    pols = ['TT','EE','TE','EB','TB']
    # est = "hdv"
    # pols = ['TT','EE','TE','ET','EB','TB']

    deg = 5.
    px = 2.0
    tellmin = 30
    tellmax = 3000
    pellmin = 30
    pellmax = 5000
    kellmin = 10
    kellmax = 5000
    bin_width = 40
    
    beam_arcmin = 1.5
    noise_uk_arcmin = 10.0
    
    theory = cosmology.default_theory(lpad=30000)
    shape,wcs = s.rect_geometry(width_deg=deg,px_res_arcmin=px)
    modlmap = enmap.modlmap(shape,wcs)
    kbeam = s.gauss_beam(modlmap,beam_arcmin)
    n2d = (noise_uk_arcmin*np.pi/180./60.)**2./kbeam**2.
    tmask = s.mask_kspace(shape,wcs,lmin=tellmin,lmax=tellmax)
    pmask = s.mask_kspace(shape,wcs,lmin=pellmin,lmax=pellmax)
    kmask = s.mask_kspace(shape,wcs,lmin=kellmin,lmax=kellmax)
    bin_edges = np.arange(kellmin,kellmax,bin_width)
    binner = s.bin2D(modlmap,bin_edges)
    
    feed_dict = {}
    cltt = theory.lCl('TT',modlmap)
    clee = theory.lCl('EE',modlmap)
    clbb = theory.lCl('BB',modlmap)
    clte = theory.lCl('TE',modlmap)
    feed_dict['uC_T_T'] = cltt
    feed_dict['tC_T_T'] = (cltt+n2d)
    feed_dict['uC_E_E'] = clee
    feed_dict['tC_E_E'] = (clee+n2d*2.)
    feed_dict['uC_B_B'] = clbb
    feed_dict['tC_B_B'] = (clbb+n2d*2.)
    feed_dict['uC_T_E'] = clte
    feed_dict['tC_T_E'] = clte

    ells = np.arange(0,10000,1)
    pl = io.Plotter(xyscale='loglog',xlabel='L',ylabel='C')
    pl.add(ells,theory.gCl('kk',ells))
    imask = {'T':tmask, 'E':pmask,'B': pmask}
    for pol in pols:
        print(pol)
        X,Y = pol
        cents,Nl = binner.bin(s.N_l(shape,wcs,feed_dict,est,pol,xmask=imask[X],ymask=imask[Y]))
        pl.add(cents,Nl,label=pol)
    pl._ax.set_xlim(10,kellmax)
    pl.done("nls.png")
    
    
